//		  Ŀ
//	> AVNER 
//	>  BEN  > Software Engineering Method
//		  
//	10 Dov-Hoz st. Tel-Aviv 63416 Israel tel. 972-3-221535

// The Screen NAVigator, ver 1.20 August 1990
// Copyright (c) 1990 by Avner Ben
// Snav is not, and never was, free software.
// for conditions for use refer to file "copyrigh.txt"

// The Screen Navigator is an object-oriented device-independent
// character-graphics driver package, written in the C++ language,
// distributed in the form of C++ source code.
// For further information refer to the documentation files.

// this simple example is intended as a template to be extended and modified
// by the user, provided the above title and copyright notice are unchanged
// and are not ommitted.

/***************************************************************************/

// demonstration part 2 - header.
// a simplistic formatted-input driver, implemented with the screen driver
// of demo part 1. Device-specific assumptions made about function keys.
// Implemented are only the methods required by the main demo program.

// 28.10.89 avner ben coded.
// 13.7.90  last update by avner ben.

// site history (this copy):
// __.__.__, _____________: _______________________________.

#ifndef DEMO1_H
#include "demo1.hpp"
#endif

enum func_k {	// function key pressed?
	K_C,	// no known function. assume character pressed!
	// function ket pressed, char=space
	K_RT,	// right arrow
	K_UP,	// up arrow
	K_LT,	// left arrow
	K_DN,	// down arrow
	K_ESC	// escape
	// add your own
};

class character_input
{ // interactive keyboard/console input driver

	private:
		boolean private_screen; // tells if delete on exit

	protected:
		panel *scr;		// specific driver determined by caller
		char cc;		// last read char. blank if fkey pressed
		func_k fkey;		// last pressed fkey. 0 if char pressed.

	public :
		character_input(panel *inscr); // user supplied window
		character_input(square_pos &window, int driver_num=0);
		virtual ~character_input(void); // create (and destroy) our own window
		virtual char get_c(void)=0; // implementation dependant
		virtual int charnum(char c)=0; // char-set dependant
		char c(void) { return cc; }
		func_k fk(void) { return fkey; }

		void clear(void) { scr->clear(); }
		void home(void) { scr->home(); }
		void move(void) { scr->move(); }
		void put_s(char *s) { scr->put_s(s); }
		void put_c_stay(char c)
		 { scr->put_c(c,NULL,&nodir); }

		// routines needed by demo prog
		boolean get_position(char *prompt, panel *draw_board, boolean hilite=TRUE);
		int get_number(char *prompt, int imax);
		boolean get_boolean(char *prompt);
		weight_d &get_wgt(void);
		direction &get_dir(void);
		void msg(char *prompt);
};

class pc_keyboard : public character_input
{
	public:
		pc_keyboard(panel *inscr) : character_input(inscr) {}
		pc_keyboard(square_pos &window, int driver_num=0)
		 : character_input(window, driver_num) {}
		char get_c(void); // the only module that does physical input!
		int charnum(char c);
};

class keyboard_driver_manager
{ // run-time keyboard-implementation selector

	private:
		int default_kbd;

	public:
		keyboard_driver_manager(int def=0)
		 { default_kbd=def; }
		character_input *allocate(panel *scr, int kbd_num=0);
		character_input *allocate(square_pos &window,
		 int scr_num=0, int kbd_num=0);
		int ask_default(void)
		 { return default_kbd; }
		void set_default(int def)
		 { default_kbd=def; }
};


struct slide_option
{ // linked-list of options in slide-menu

	char *s,						// option name
	 *letter,				// option letter in the former
	 *hlp;				       // explanation (in dialog area)
	int num;				     // option number (serial)
	point_pos pos;				      // position in menu area
	slide_option *suc, *prd;

	slide_option(char *text, int posopt, char *help);
	~slide_option(void)
	 { delete s; delete hlp; }
	void off(panel *scr)
	 {
		scr->toggle_attr(VD_HI);
		scr->put_s(s,&pos); scr->toggle_attr(VD_HI);
	 }
	void on(panel *scr)	    // pc-specific visual attribute codes used
	 {
		scr->toggle_attr(VD_REV);
		scr->put_s(s,&pos); scr->toggle_attr(VD_REV);
	 }
	void put_help(character_input *explain)
	 { explain->clear(); explain->put_s(hlp); }
};

class slide_menu
{ // slide-menu manager (like pop-up menu system topmost menu)

	friend struct slide_option;

	private :
		panel *scr;	// the menu area, the help area
		character_input *explain;
		slide_option *stopt, *enopt, *cursor; // ptrs to option list

	public :
		slide_menu(panel *parent, character_input *help);
		 // warning: chops top line off parent screen!
		~slide_menu(void);
		void append_opt(char *text, int posopt, char *help);
		void succ_opt(void);
		void pred_opt(void);
		int select(void);
		void clear(void)
		 { scr->clear(); }
};

class text_window
{
	private :
		panel *box;
		character_input *kbd;

	public:
		text_window(panel *parent, point_pos *start, char *title,
		 boolean centered, char *inpline[]);
		~text_window(void)
		 { delete box; }
		void pause(char *prompt);
		void put_page(char *inpline[]);
};

